import bpy
from math import radians

#Props to ivan2340 at blenderartist.org... I used some of his ragdoll to put this together
def main():
    """
    Adds ragdoll pieces to target bone locations, rotations, and scales.
    Adds rigid body joint constraints to ragdoll pieces with default values.
    Adds Follow objects to the head of each of the target bones, parenting them to their corresponding bones
    Adds copy rotation constraints to the bones called 'RAGDOLL_ROTATION'
    Adds copy location constraint to the hips bone called 'RAGDOLL_LOCATION'
    Instructions: Select metarig, ensure bone names are correct, run script.
    """
    
    rig = bpy.context.scene.objects.active
    rig.select = False
    
    RD_PREFIX = 'RD-'
    FOLLOW_PREFIX = 'FOLLOW-'
    
    BODY_MASS = 100
    
    head = 'head'
    neck = 'neck'
    chest = 'chest'
    spine = 'spine'
    hips = 'hips'
    shoulder_L = 'shoulder.L'
    shoulder_R = 'shoulder.R'
    upper_arm_L = 'upper_arm.L'
    upper_arm_R = 'upper_arm.R'
    forearm_L = 'forearm.L'
    forearm_R = 'forearm.R'
    hand_L = 'hand.L'
    hand_R = 'hand.R'
    thigh_L = 'thigh.L'
    thigh_R = 'thigh.R'
    shin_L = 'shin.L'
    shin_R = 'shin.R'
    foot_L = 'foot.L'
    foot_R = 'foot.R'
    toe_L = 'toe.L'
    toe_R = 'toe.R'
    
    bone_name_list = [head, neck, chest, spine, hips, shoulder_L, shoulder_R, upper_arm_L, upper_arm_R, forearm_L, forearm_R, hand_L, hand_R, thigh_L, thigh_R, shin_L, shin_R, foot_L, foot_R, toe_L, toe_R]
    
    finger_list = ['f_index.01.L', 'f_index.02.L', 'f_index.03.L', 'thumb.01.L', 'thumb.02.L', 'thumb.03.L', 'f_middle.01.L', 'f_middle.02.L', 'f_middle.03.L', 'f_ring.01.L', 'f_ring.02.L', 'f_ring.03.L', 'f_pinky.01.L', 'f_pinky.02.L', 'f_pinky.03.L', 'f_index.01.R', 'f_index.02.R', 'f_index.03.R', 'thumb.01.R', 'thumb.02.R', 'thumb.03.R', 'f_middle.01.R', 'f_middle.02.R', 'f_middle.03.R', 'palm.03.R', 'f_ring.01.R', 'f_ring.02.R', 'f_ring.03.R', 'f_pinky.01.R', 'f_pinky.02.R', 'f_pinky.03.R']
     
    create_ragdoll(rig, RD_PREFIX, BODY_MASS, bone_name_list, finger_list)
    #create_ragdoll_widgets(rig, bone_name_list, FOLLOW_PREFIX, RD_PREFIX)
    #add_bone_constraints(rig, RD_PREFIX, bone_name_list)
    #Make Relevant Layers visible 
    #bpy.context.scene.layers = create_layers_boolean([(bpy.context.scene.active_layer + 1), 19], 20)
    
def create_layers_boolean(list_of_layers, total_layers):
    """
    Given a list of layers and the total number of layers
    returns a boolean list of Trues and False
    """
    boolean_list = []
    for i in range(1, (total_layers + 1)):
        if i in list_of_layers:
            boolean_list.append(True)
        else:
            boolean_list.append(False)
    return boolean_list

def create_body_part(rig, bone_name, RD_PREFIX):
        """ Creates an empty object for a ragdoll body part, and returns the object.
        """
        obj_name = RD_PREFIX + bone_name
        scene = bpy.context.scene
        # Check if it already exists
        if obj_name in scene.objects:
            return None
        else:
            mesh = bpy.data.meshes.new(obj_name)
            obj = bpy.data.objects.new(obj_name, mesh)
            scene.objects.link(obj)
    
            obj_to_bone(obj, rig, bone_name, 'head', False)
    
            return obj

def obj_to_bone(obj, rig, bone_name, where, make_bone_parent):
        """ 
        Places an object at the location/rotation/scale of the given bone Warning! This script will clear
        whatever object was the parent of the obj before running the script.
        
        Parameters:     obj (bpy.types.Objects) the object that will be moved to the bone's position
                    
                        rig (bpy.types.Objects) the rig that has the bone to which the object will be moved
                        
                        bone_name (string) the name of the bone to which the object will be moved
                        
                        where (string) the location along the bone where the object should be placed 'head' 'tail' or 'middle' 
                        
                        make_bone_parent (boolean) Whether or not the object should be parented to the bone after being placed at
                        the bone's location.
        """       
        obj.select = True
        bpy.ops.object.parent_clear(type='CLEAR')
        bpy.ops.object.location_clear()
        bpy.ops.object.rotation_clear()
        bpy.ops.object.scale_clear()        
        obj.parent = rig
        obj.parent_type = 'BONE'
        obj.parent_bone = bone_name
        bone = rig.data.bones[bone_name]        
        if where == 'head':
            obj.location[1] -= bone.length
        elif where == 'middle':
            obj.location[1] -= bone.length/2        
        obj.rotation_euler = obj.parent.rotation_euler
        if make_bone_parent == True:
            obj.rotation_euler = obj.parent.rotation_euler
        else:
            bpy.ops.object.parent_clear(type='CLEAR_KEEP_TRANSFORM')
            mat = rig.matrix_world * bone.matrix_local
            obj.rotation_mode = ('XYZ')
            obj.rotation_euler = mat.to_euler()    
            scl = mat.to_scale()
            scl_avg = (scl[0] + scl[1] + scl[2]) / 3
            obj.scale = (bone.length * scl_avg), (bone.length * scl_avg), (bone.length * scl_avg)
        obj.select = False
            

            
        
#sets some standard things like physics type, mass etc. "ratio" is the proportion of mass given to each body part.
def standard_doll_settings(obj, ratio, name, BODY_MASS, RD_PREFIX):
    if obj != None:
        obj.game.physics_type = 'RIGID_BODY'
        obj.hide_render = False
        obj.game.mass = (ratio * BODY_MASS)
        obj.game.use_collision_bounds = True
        obj.game.collision_bounds_type = 'CONVEX_HULL'
        obj.select = True
        bpy.ops.object.transform_apply(location=False, rotation=False, scale=True)
        obj.select = False

    else:
        obj = bpy.data.objects[RD_PREFIX + name]

def create_ragdoll(rig, RD_PREFIX, BODY_MASS, bone_name_list, finger_list): 
    """ 
    Generates a ragdoll base using some of the bones of a metarig armature
        
    Parameters:     rig (bpy.data.objects) The armature object that will be recieving the ragdoll pieces
                    
                    RD_PREFIX (string) the prefix that will be added to each of the generated ragdoll
                    pieces
                    
                    layer (int between 1 and 20) the layer to which all of the ragdoll pieces will 
                    be added
                    
                    BODY_MASS (float) the total body mass of the ragdoll. Each piece is given a portion
    """
            
    #=============================================
    # BODY PARTS
    #=============================================
    
    def create_head(rig, bone_name, RD_PREFIX):
        obj = create_body_part(rig, bone_name, RD_PREFIX)
        if obj != None:
            verts = [(0.0, 0.57744, -0.32974), (-0.54717, 0.53556, 0.27016), (0.0, 0.49654, 0.83042), (0.54717, 0.53556, 0.27016), (0.0, 1.16007, 0.29682), (0.0, 0.02924, 0.28585), (-0.35191, 0.57217, -0.25459), (-0.35179, 0.49654, 0.83042), (0.35191, 0.57217, -0.25459), (0.35179, 0.49654, 0.83042), (0.0, 1.02601, -0.12757), (-0.47948, 1.00959, 0.28634), (0.0, 1.02414, 0.67698), (0.47948, 1.00959, 0.28634), (-0.35204, 0.02924, 0.28585), (-0.0, 0.05542, -0.16296), (0.35204, 0.02924, 0.28585), (0.0, -0.1819, 0.76109), (-0.35191, 0.97287, -0.04249), (-0.35191, 0.97105, 0.5845), (0.35191, 0.97105, 0.5845), (0.35191, 0.97287, -0.04249), (-0.31458, 0.05542, -0.16296), (-0.25041, -0.1819, 0.76109), (0.25041, -0.1819, 0.76109), (0.31458, 0.05542, -0.16296)]
            faces = [(10, 0, 6, 18), (0, 15, 22, 6), (0, 10, 21, 8), (15, 0, 8, 25), (14, 1, 6, 22), (1, 11, 18, 6), (23, 7, 1, 14), (7, 19, 11, 1), (17, 2, 7, 23), (2, 12, 19, 7), (24, 9, 2, 17), (9, 20, 12, 2), (16, 3, 9, 24), (3, 13, 20, 9), (25, 8, 3, 16), (8, 21, 13, 3), (13, 4, 12, 20), (4, 11, 19, 12), (21, 10, 4, 13), (10, 18, 11, 4), (15, 5, 14, 22), (5, 17, 23, 14), (25, 16, 5, 15), (16, 24, 17, 5)]
            mesh = obj.data
            mesh.from_pydata(verts, [], faces)
            mesh.update() 
            return obj
        else:
            return None
        
    def create_neck(rig, bone_name, RD_PREFIX):
        obj = create_body_part(rig, bone_name, RD_PREFIX)
        if obj != None:
            verts =[(-0.30853, 0.81944, -0.30097), (-0.30853, 0.81953, 0.32071), (0.30853, 0.81953, 0.32071), (0.30853, 0.81944, -0.30097), (-0.30853, 0.10838, -0.30094), (-0.30853, 0.10848, 0.32075), (0.30853, 0.10848, 0.32075), (0.30853, 0.10838, -0.30094)]
            faces = [(4, 0, 3, 7), (0, 4, 5, 1), (6, 2, 1, 5), (7, 3, 2, 6), (3, 0, 1, 2), (7, 6, 5, 4)]
            mesh = obj.data
            mesh.from_pydata(verts, [], faces)
            mesh.update() 
            return obj
        else:
            return None
    
    def create_chest(rig, bone_name, RD_PREFIX):
        obj = create_body_part(rig, bone_name, RD_PREFIX)
        if obj != None:
            verts = [(-0.33506, 0.95883, -0.31085), (-0.33506, 0.97051, 0.30445), (0.33506, 0.97051, 0.30445), (0.33506, 0.95883, -0.31085), (-0.28525, 0.09357, -0.24828), (-0.3532, 0.10856, 0.35624), (0.3532, 0.10856, 0.35624), (0.28525, 0.09357, -0.24828), (0.36553, 0.74996, -0.32931), (0.38523, 0.38203, -0.30187), (-0.36553, 0.78864, 0.40879), (-0.43663, 0.43811, 0.37317), (-0.36553, 0.74996, -0.32931), (-0.38523, 0.38203, -0.30187), (0.36553, 0.78864, 0.40879), (0.43663, 0.43811, 0.37317)]
            faces = [(12, 0, 3, 8), (13, 4, 5, 11), (14, 2, 1, 10), (8, 3, 2, 14), (3, 0, 1, 2), (7, 6, 5, 4), (4, 13, 9, 7), (13, 12, 8, 9), (0, 12, 10, 1), (12, 13, 11, 10), (6, 15, 11, 5), (15, 14, 10, 11), (7, 9, 15, 6), (9, 8, 14, 15)]
            mesh = obj.data
            mesh.from_pydata(verts, [], faces)
            mesh.update() 
            return obj
        else:
            return None
        
    def create_shoulder_L(rig, bone_name, RD_PREFIX):
        obj = create_body_part(rig, bone_name, RD_PREFIX)
        if obj != None:
            verts = [(0.1177, 0.67809, -0.19664), (-0.40493, 0.98559, -0.1965), (-0.34117, 1.09186, -0.19643), (0.18147, 0.78436, -0.19657), (0.09735, 0.64032, 0.25865), (-0.42528, 0.94782, 0.25878), (-0.22576, 1.28451, 0.22492), (0.29687, 0.97701, 0.22479)]
            faces = [(4, 5, 1, 0), (5, 6, 2, 1), (6, 7, 3, 2), (7, 4, 0, 3), (0, 1, 2, 3), (7, 6, 5, 4)]
            mesh = obj.data
            mesh.from_pydata(verts, [], faces)
            mesh.update() 
            return obj
        else:
            return None
        
    def create_shoulder_R(rig, bone_name, RD_PREFIX):
        obj = create_body_part(rig, bone_name, RD_PREFIX)
        if obj != None:
            verts = [(-0.1177, 0.67809, -0.19664), (0.40493, 0.98559, -0.1965), (0.34117, 1.09186, -0.19643), (-0.18147, 0.78436, -0.19657), (-0.09735, 0.64032, 0.25865), (0.42528, 0.94782, 0.25878), (0.22576, 1.28451, 0.22492), (-0.29687, 0.97701, 0.22479)]
            faces = [(4, 5, 1, 0), (5, 6, 2, 1), (6, 7, 3, 2), (7, 4, 0, 3), (0, 1, 2, 3), (7, 6, 5, 4)]
            mesh = obj.data
            mesh.from_pydata(verts, [], faces)
            mesh.update() 
            return obj
        else:
            return None
        
    def create_upper_arm_L(rig, bone_name, RD_PREFIX):
        obj = create_body_part(rig, bone_name, RD_PREFIX)
        if obj != None:
            verts = [(0.16749, 0.06784, 0.09715), (0.16749, 0.0, -0.16363), (0.16749, 0.9772, -0.16363), (0.16749, 0.9772, 0.00514), (-0.16749, 0.06784, 0.09715), (-0.16749, 0.0, -0.16363), (-0.16749, 0.9772, -0.16363), (-0.16749, 0.9772, 0.00514), (-0.16749, 0.65291, -0.16363), (0.16749, 0.65291, -0.16363), (-0.16749, 0.65291, 0.16363), (0.16749, 0.65291, 0.16363)]
            faces = [(4, 5, 1, 0), (8, 6, 2, 9), (6, 7, 3, 2), (10, 4, 0, 11), (11, 9, 2, 3), (10, 8, 5, 4), (5, 8, 9, 1), (7, 10, 11, 3), (0, 1, 9, 11), (7, 6, 8, 10)]
            mesh = obj.data
            mesh.from_pydata(verts, [], faces)
            mesh.update() 
            return obj
        else:
            return None
        
    def create_upper_arm_R(rig, bone_name, RD_PREFIX):
        obj = create_body_part(rig, bone_name, RD_PREFIX)
        if obj != None:
            verts = [(0.16749, 0.06784, 0.09715), (0.16749, 0.0, -0.16363), (0.16749, 0.9772, -0.16363), (0.16749, 0.9772, 0.00514), (-0.16749, 0.06784, 0.09715), (-0.16749, 0.0, -0.16363), (-0.16749, 0.9772, -0.16363), (-0.16749, 0.9772, 0.00514), (-0.16749, 0.65291, -0.16363), (0.16749, 0.65291, -0.16363), (-0.16749, 0.65291, 0.16363), (0.16749, 0.65291, 0.16363)]
            faces = [(4, 5, 1, 0), (8, 6, 2, 9), (6, 7, 3, 2), (10, 4, 0, 11), (11, 9, 2, 3), (10, 8, 5, 4), (5, 8, 9, 1), (7, 10, 11, 3), (0, 1, 9, 11), (7, 6, 8, 10)]
            mesh = obj.data
            mesh.from_pydata(verts, [], faces)
            mesh.update() 
            return obj
        else:
            return None
        
    def create_forearm_L(rig, bone_name, RD_PREFIX):
        obj = create_body_part(rig, bone_name, RD_PREFIX)
        if obj != None:
            verts = [(0.10073, 0.99095, 0.12967), (0.10073, 0.99095, -0.11157), (0.17114, 0.01375, -0.19589), (0.17114, 0.01375, 0.01549), (-0.10073, 0.99095, 0.12967), (-0.10073, 0.99095, -0.11157), (-0.17114, 0.01375, -0.19589), (-0.17114, 0.01375, 0.01549), (-0.17114, 0.26733, -0.19589), (0.17114, 0.26733, -0.19589), (-0.17114, 0.26733, 0.14313), (0.17114, 0.26733, 0.14313)]
            faces = [(4, 5, 1, 0), (8, 6, 2, 9), (6, 7, 3, 2), (10, 4, 0, 11), (11, 9, 2, 3), (10, 8, 5, 4), (5, 8, 9, 1), (7, 10, 11, 3), (0, 1, 9, 11), (7, 6, 8, 10)]
            mesh = obj.data
            mesh.from_pydata(verts, [], faces)
            mesh.update() 
            return obj
        else:
            return None
        
    def create_forearm_R(rig, bone_name, RD_PREFIX):
        obj = create_body_part(rig, bone_name, RD_PREFIX)
        if obj != None:
            verts = [(0.10073, 0.99095, 0.12967), (0.10073, 0.99095, -0.11157), (0.17114, 0.01375, -0.19589), (0.17114, 0.01375, 0.01549), (-0.10073, 0.99095, 0.12967), (-0.10073, 0.99095, -0.11157), (-0.17114, 0.01375, -0.19589), (-0.17114, 0.01375, 0.01549), (-0.17114, 0.26733, -0.19589), (0.17114, 0.26733, -0.19589), (-0.17114, 0.26733, 0.14313), (0.17114, 0.26733, 0.14313)]
            faces = [(4, 5, 1, 0), (8, 6, 2, 9), (6, 7, 3, 2), (10, 4, 0, 11), (11, 9, 2, 3), (10, 8, 5, 4), (5, 8, 9, 1), (7, 10, 11, 3), (0, 1, 9, 11), (7, 6, 8, 10)]
            mesh = obj.data
            mesh.from_pydata(verts, [], faces)
            mesh.update() 
            return obj
        else:
            return None
        
    def create_hand_L(rig, bone_name, RD_PREFIX):
        obj = create_body_part(rig, bone_name, RD_PREFIX)
        if obj != None:
            verts = [(-0.41019, 0.10352, -0.20714), (-0.3923, 0.10352, -0.00763), (0.3571, 0.10352, -0.20714), (0.37499, 0.10352, -0.00763), (0.59157, 0.58064, -0.20714), (-0.61481, 0.58064, -0.20714), (0.59157, 0.58064, 0.17325), (-0.61481, 0.58064, 0.17325), (-0.61481, 1.30177, 0.17325), (0.59157, 1.30177, 0.17325), (-0.61481, 1.30177, -0.20714), (0.59157, 1.30177, -0.20714)]
            faces = [(4, 2, 0, 5), (2, 3, 1, 0), (7, 5, 0, 1), (3, 6, 7, 1), (3, 2, 4, 6), (4, 5, 10, 11), (5, 7, 8, 10), (6, 4, 11, 9), (7, 6, 9, 8), (10, 8, 9, 11)]
            mesh = obj.data
            mesh.from_pydata(verts, [], faces)
            mesh.update() 
            return obj
        else:
            return None
        
    def create_hand_R(rig, bone_name, RD_PREFIX):
        obj = create_body_part(rig, bone_name, RD_PREFIX)
        if obj != None:
            verts = [(0.41019, 0.10352, -0.20714), (0.3923, 0.10352, -0.00763), (-0.3571, 0.10352, -0.20714), (-0.37499, 0.10352, -0.00763), (-0.59157, 0.58064, -0.20714), (0.61481, 0.58064, -0.20714), (-0.59157, 0.58064, 0.17325), (0.61481, 0.58064, 0.17325), (0.61481, 1.30177, 0.17325), (-0.59157, 1.30177, 0.17325), (0.61481, 1.30177, -0.20714), (-0.59157, 1.30177, -0.20714)]
            faces = [(4, 2, 0, 5), (2, 3, 1, 0), (7, 5, 0, 1), (3, 6, 7, 1), (3, 2, 4, 6), (4, 5, 10, 11), (5, 7, 8, 10), (6, 4, 11, 9), (7, 6, 9, 8), (10, 8, 9, 11)]
            mesh = obj.data
            mesh.from_pydata(verts, [], faces)
            mesh.update() 
            return obj
        else:
            return None
        
    def create_spine(rig, bone_name, RD_PREFIX):
        obj = create_body_part(rig, bone_name, RD_PREFIX)
        if obj != None:
            verts = [(-0.69515, 1.06281, 0.57129), (-0.85384, 0.40997, 0.79654), (0.85384, 0.40997, 0.79654), (0.69515, 1.06281, 0.57129), (-0.61707, 0.96972, -0.48418), (-0.75793, 0.25929, -0.62154), (0.75793, 0.25929, -0.62154), (0.61707, 0.96972, -0.48418)]
            faces = [(4, 0, 3, 7), (0, 4, 5, 1), (6, 2, 1, 5), (7, 3, 2, 6), (3, 0, 1, 2), (7, 6, 5, 4)]
            mesh = obj.data
            mesh.from_pydata(verts, [], faces)
            mesh.update() 
            return obj
        else:
            return None
        
    def create_hips(rig, bone_name, RD_PREFIX):
        obj = create_body_part(rig, bone_name, RD_PREFIX)
        if obj != None:
            verts =[(0.69408, 1.0016, -0.49999), (0.69408, 1.22485, 0.65103), (-0.69408, 1.22485, 0.65103), (-0.69408, 1.0016, -0.49999), (0.77408, 0.72683, 0.64196), (-0.63944, 0.42863, -0.46287), (-0.77408, 0.72683, 0.64196), (0.63944, 0.42863, -0.46287)]
            faces = [(7, 0, 1, 4), (3, 0, 7, 5), (1, 2, 6, 4), (4, 6, 5, 7), (3, 5, 6, 2), (0, 3, 2, 1)]
            mesh = obj.data
            mesh.from_pydata(verts, [], faces)
            mesh.update() 
            return obj
        else:
            return None
        
    def create_thigh_L(rig, bone_name, RD_PREFIX):
        obj = create_body_part(rig, bone_name, RD_PREFIX)
        if obj != None:
            verts =[(-0.16692, 0.44145, 0.02371), (0.16141, 0.44145, 0.02371), (-0.09494, 0.0, 0.0), (0.09494, 0.0, 0.0), (-0.13663, 0.95579, 0.09109), (0.13142, 0.95579, 0.09109), (-0.11398, 0.40816, 0.20367), (-0.11398, 0.47471, -0.15624), (0.10843, 0.40816, 0.20367), (0.10843, 0.47471, -0.15624), (-0.09404, 0.10283, 0.20749), (-0.09404, 0.32194, -0.17924), (0.08849, 0.32194, -0.17924), (0.08849, 0.10283, 0.20749), (-0.08227, 0.75197, 0.14277), (-0.09339, 1.07879, -0.05811), (0.0882, 1.07879, -0.05811), (0.07676, 0.75197, 0.14277)]
            faces = [(2, 0, 6, 10), (0, 2, 11, 7), (0, 4, 14, 6), (4, 0, 7, 15), (3, 1, 9, 12), (13, 8, 1, 3), (1, 5, 16, 9), (8, 17, 5, 1), (2, 3, 12, 11), (10, 13, 3, 2), (5, 4, 15, 16), (17, 14, 4, 5), (14, 17, 8, 6), (13, 10, 6, 8), (9, 7, 11, 12), (16, 15, 7, 9)]
            mesh = obj.data
            mesh.from_pydata(verts, [], faces)
            mesh.update() 
            return obj
        else:
            return None
        
    def create_thigh_R(rig, bone_name, RD_PREFIX):
        obj = create_body_part(rig, bone_name, RD_PREFIX)
        if obj != None:
            verts = [(0.16412, 0.44145, 0.02371), (-0.16421, 0.44145, 0.02371), (0.09214, 0.0, 0.0), (-0.09774, 0.0, 0.0), (0.13383, 0.95579, 0.09109), (-0.13422, 0.95579, 0.09109), (0.11118, 0.40816, 0.20367), (0.11118, 0.47471, -0.15624), (-0.11123, 0.40816, 0.20367), (-0.11123, 0.47471, -0.15624), (0.09124, 0.10283, 0.20749), (0.09124, 0.32194, -0.17924), (-0.09129, 0.32194, -0.17924), (-0.09129, 0.10283, 0.20749), (0.07947, 0.75197, 0.14277), (0.09059, 1.07879, -0.05811), (-0.091, 1.07879, -0.05811), (-0.07956, 0.75197, 0.14277)]
            faces = [(2, 0, 6, 10), (0, 2, 11, 7), (0, 4, 14, 6), (4, 0, 7, 15), (3, 1, 9, 12), (13, 8, 1, 3), (1, 5, 16, 9), (8, 17, 5, 1), (2, 3, 12, 11), (10, 13, 3, 2), (5, 4, 15, 16), (17, 14, 4, 5), (14, 17, 8, 6), (13, 10, 6, 8), (9, 7, 11, 12), (16, 15, 7, 9)]
            mesh = obj.data
            mesh.from_pydata(verts, [], faces)
            mesh.update() 
            return obj
        else:
            return None
        
    def create_shin_L(rig, bone_name, RD_PREFIX):
        obj = create_body_part(rig, bone_name, RD_PREFIX)
        if obj != None:
            verts = [(0.10691, 0.09779, 0.05968), (-0.10882, 0.09779, 0.05968), (0.07341, 0.90035, 0.05418), (-0.0753, 0.90035, 0.05418), (0.10691, 0.43809, 0.17558), (0.10691, 0.1044, -0.08664), (-0.10882, 0.1044, -0.08664), (-0.10882, 0.43809, 0.17558), (0.07341, 0.90794, 0.14516), (0.07341, 0.89277, -0.03681), (-0.0753, 0.89277, -0.03681), (-0.0753, 0.90794, 0.14516)]
            faces = [(0, 2, 8, 4), (2, 0, 5, 9), (1, 3, 10, 6), (7, 11, 3, 1), (0, 1, 6, 5), (1, 0, 4, 7), (3, 2, 9, 10), (11, 8, 2, 3), (8, 11, 7, 4), (10, 9, 5, 6)]
            mesh = obj.data
            mesh.from_pydata(verts, [], faces)
            mesh.update() 
            return obj
        else:
            return None
        
    def create_shin_R(rig, bone_name, RD_PREFIX):
        obj = create_body_part(rig, bone_name, RD_PREFIX)
        if obj != None:
            verts =[(-0.10716, 0.09813, 0.05964), (0.10906, 0.09813, 0.05964), (-0.07358, 0.90248, 0.05258), (0.07548, 0.90248, 0.05258), (-0.10716, 0.43942, 0.17515), (-0.10716, 0.10448, -0.08704), (0.10906, 0.10448, -0.08704), (0.10906, 0.43942, 0.17515), (-0.07358, 0.91026, 0.14376), (-0.07358, 0.89471, -0.03859), (0.07548, 0.89471, -0.03859), (0.07548, 0.91026, 0.14376)]
            faces = [(0, 2, 8, 4), (2, 0, 5, 9), (1, 3, 10, 6), (7, 11, 3, 1), (0, 1, 6, 5), (1, 0, 4, 7), (3, 2, 9, 10), (11, 8, 2, 3), (8, 11, 7, 4), (10, 9, 5, 6)]
            mesh = obj.data
            mesh.from_pydata(verts, [], faces)
            mesh.update() 
            return obj
        else:
            return None
        
    def create_foot_L(rig, bone_name, RD_PREFIX):
        obj = create_body_part(rig, bone_name, RD_PREFIX)
        if obj != None:
            verts = [(-0.28227, -0.40496, 0.02862), (-0.30662, -0.03346, 0.79829), (0.28227, -0.40496, 0.02862), (0.30662, -0.03346, 0.79829), (0.28227, 0.13894, -0.2349), (0.38141, 0.92782, -0.05066), (-0.38141, 0.92782, -0.05066), (-0.28227, 0.13894, -0.2349), (0.38141, 0.64471, 0.37443), (0.38141, 1.03847, 0.12833), (-0.38141, 1.03847, 0.12833), (-0.38141, 0.64471, 0.37443)]
            faces = [(0, 2, 3, 1), (11, 1, 3, 8), (0, 7, 4, 2), (7, 0, 1, 11), (8, 3, 2, 4), (9, 8, 4, 5), (10, 9, 5, 6), (11, 10, 6, 7), (7, 6, 5, 4), (11, 8, 9, 10)]
            mesh = obj.data
            mesh.from_pydata(verts, [], faces)
            mesh.update() 
            return obj
        else:
            return None
        
    def create_foot_R(rig, bone_name, RD_PREFIX):
        obj = create_body_part(rig, bone_name, RD_PREFIX)
        if obj != None:
            verts = [(0.28227, -0.40496, 0.02862), (0.30662, -0.03346, 0.79829), (-0.28227, -0.40496, 0.02862), (-0.30662, -0.03346, 0.79829), (-0.28227, 0.13894, -0.2349), (-0.38141, 0.92782, -0.05066), (0.38141, 0.92782, -0.05066), (0.28227, 0.13894, -0.2349), (-0.38141, 0.64471, 0.37443), (-0.38141, 1.03847, 0.12833), (0.38141, 1.03847, 0.12833), (0.38141, 0.64471, 0.37443)]
            faces = [(0, 2, 3, 1), (11, 1, 3, 8), (0, 7, 4, 2), (7, 0, 1, 11), (8, 3, 2, 4), (9, 8, 4, 5), (10, 9, 5, 6), (11, 10, 6, 7), (7, 6, 5, 4), (11, 8, 9, 10)]
            mesh = obj.data
            mesh.from_pydata(verts, [], faces)
            mesh.update() 
            return obj
        else:
            return None
        
    def create_toe_L(rig, bone_name, RD_PREFIX):
        obj = create_body_part(rig, bone_name, RD_PREFIX)
        if obj != None:
            verts = [(-0.65942, 0.06729, -0.23464), (-0.65942, 0.06729, 0.20231), (0.65942, 0.06729, -0.23464), (0.65942, 0.06729, 0.20231), (0.65942, 1.09674, -0.23464), (-0.65942, 1.09674, -0.23464), (0.65942, 1.09674, 0.20231), (-0.65942, 1.09674, 0.20231)]
            faces = [(0, 2, 3, 1), (7, 1, 3, 6), (0, 5, 4, 2), (5, 0, 1, 7), (6, 3, 2, 4), (4, 5, 7, 6)]
            mesh = obj.data
            mesh.from_pydata(verts, [], faces)
            mesh.update() 
            return obj
        else:
            return None
        
    def create_toe_R(rig, bone_name, RD_PREFIX):
        obj = create_body_part(rig, bone_name, RD_PREFIX)
        if obj != None:
            verts = [(3.57609, 0.06729, -0.23464), (3.57609, 0.06729, 0.20231), (2.25725, 0.06729, -0.23464), (2.25725, 0.06729, 0.20231), (2.25725, 1.09674, -0.23464), (3.57609, 1.09674, -0.23464), (2.25725, 1.09674, 0.20231), (3.57609, 1.09674, 0.20231)]
            faces = [(0, 2, 3, 1), (7, 1, 3, 6), (0, 5, 4, 2), (5, 0, 1, 7), (6, 3, 2, 4), (4, 5, 7, 6)]
            mesh = obj.data
            mesh.from_pydata(verts, [], faces)
            mesh.update() 
            return obj
        else:
            return None
        
    def create_finger(rig, bone_name, RD_PREFIX):
        obj = create_body_part(rig, bone_name, RD_PREFIX)
        if obj != None:
            verts = [(0.1767, 0.1626, -0.1971), (0.1767, 0.1626, 0.12687), (-0.1484, 0.1626, -0.1971), (-0.1484, 0.1626, 0.12687), (-0.1484, 1.09991, -0.1971), (0.1767, 1.09991, -0.1971), (-0.1484, 0.95035, 0.12687), (0.1767, 0.95035, 0.12687), (0.1767, 0.64487, -0.1971), (-0.1484, 0.64487, -0.1971), (-0.1484, 0.57989, 0.12687), (0.1767, 0.57989, 0.12687)]
            faces = [(0, 2, 3, 1), (11, 1, 3, 10), (8, 5, 4, 9), (8, 0, 1, 11), (10, 3, 2, 9), (7, 11, 10, 6), (0, 8, 9, 2), (5, 8, 11, 7), (6, 10, 9, 4), (4, 5, 7, 6)]
            mesh = obj.data
            mesh.from_pydata(verts, [], faces)
            mesh.update() 
            return obj
        else:
            return None
       
    #Create body parts:    
    standard_doll_settings(create_head(rig, bone_name_list[0], RD_PREFIX), .07, bone_name_list[0], BODY_MASS, RD_PREFIX)
    standard_doll_settings(create_neck(rig, bone_name_list[1], RD_PREFIX), .03, bone_name_list[1], BODY_MASS, RD_PREFIX)
    standard_doll_settings(create_chest(rig, bone_name_list[2], RD_PREFIX), .2, bone_name_list[2], BODY_MASS, RD_PREFIX)
    standard_doll_settings(create_spine(rig, bone_name_list[3], RD_PREFIX), .1, bone_name_list[3], BODY_MASS, RD_PREFIX)
    standard_doll_settings(create_hips(rig, bone_name_list[4], RD_PREFIX), .2, bone_name_list[4], BODY_MASS, RD_PREFIX)  
    
    standard_doll_settings(create_shoulder_L(rig, bone_name_list[5], RD_PREFIX), .01, bone_name_list[5], BODY_MASS, RD_PREFIX)
    standard_doll_settings(create_shoulder_R(rig, bone_name_list[6], RD_PREFIX), .01, bone_name_list[6], BODY_MASS, RD_PREFIX)
    standard_doll_settings(create_upper_arm_L(rig, bone_name_list[7], RD_PREFIX), .03, bone_name_list[7], BODY_MASS, RD_PREFIX)
    standard_doll_settings(create_upper_arm_R(rig, bone_name_list[8], RD_PREFIX), .03, bone_name_list[8], BODY_MASS, RD_PREFIX)
    standard_doll_settings(create_forearm_L(rig, bone_name_list[9], RD_PREFIX), .03, bone_name_list[9], BODY_MASS, RD_PREFIX)
    standard_doll_settings(create_forearm_R(rig, bone_name_list[10], RD_PREFIX), .03, bone_name_list[10], BODY_MASS, RD_PREFIX)
    standard_doll_settings(create_hand_L(rig, bone_name_list[11], RD_PREFIX), .01, bone_name_list[11], BODY_MASS, RD_PREFIX)
    standard_doll_settings(create_hand_R(rig, bone_name_list[12], RD_PREFIX), .01, bone_name_list[12], BODY_MASS, RD_PREFIX)    
    
    standard_doll_settings(create_thigh_L(rig, bone_name_list[13], RD_PREFIX), .1, bone_name_list[13], BODY_MASS, RD_PREFIX)
    standard_doll_settings(create_thigh_R(rig, bone_name_list[14], RD_PREFIX), .1, bone_name_list[14], BODY_MASS, RD_PREFIX)
    standard_doll_settings(create_shin_L(rig, bone_name_list[15], RD_PREFIX), .05, bone_name_list[15], BODY_MASS, RD_PREFIX)
    standard_doll_settings(create_shin_R(rig, bone_name_list[16], RD_PREFIX), .05, bone_name_list[16], BODY_MASS, RD_PREFIX)
    standard_doll_settings(create_foot_L(rig, bone_name_list[17], RD_PREFIX), .01, bone_name_list[17], BODY_MASS, RD_PREFIX)
    standard_doll_settings(create_foot_R(rig, bone_name_list[18], RD_PREFIX), .01, bone_name_list[18], BODY_MASS, RD_PREFIX)
    standard_doll_settings(create_toe_L(rig, bone_name_list[19], RD_PREFIX), 0, bone_name_list[19], BODY_MASS, RD_PREFIX) 
    standard_doll_settings(create_toe_L(rig, bone_name_list[20], RD_PREFIX), 0, bone_name_list[20], BODY_MASS, RD_PREFIX)
    
    for i in range(len(finger_list)):
        standard_doll_settings(create_finger(rig, finger_list[i], RD_PREFIX), 0, finger_list[i], BODY_MASS, RD_PREFIX)
        
    
    head = bpy.data.objects[RD_PREFIX + bone_name_list[0]]
    neck = bpy.data.objects[RD_PREFIX + bone_name_list[1]]
    chest = bpy.data.objects[RD_PREFIX + bone_name_list[2]]
    spine = bpy.data.objects[RD_PREFIX + bone_name_list[3]]
    hips = bpy.data.objects[RD_PREFIX + bone_name_list[4]]
    shoulder_L = bpy.data.objects[RD_PREFIX + bone_name_list[5]]
    shoulder_R = bpy.data.objects[RD_PREFIX + bone_name_list[6]]
    upper_arm_L = bpy.data.objects[RD_PREFIX + bone_name_list[7]]
    upper_arm_R = bpy.data.objects[RD_PREFIX + bone_name_list[8]]
    forearm_L = bpy.data.objects[RD_PREFIX + bone_name_list[9]]
    forearm_R = bpy.data.objects[RD_PREFIX + bone_name_list[10]]
    hand_L = bpy.data.objects[RD_PREFIX + bone_name_list[11]] 
    hand_R = bpy.data.objects[RD_PREFIX + bone_name_list[12]]
    thigh_L = bpy.data.objects[RD_PREFIX + bone_name_list[13]]
    thigh_R = bpy.data.objects[RD_PREFIX + bone_name_list[14]]
    shin_L = bpy.data.objects[RD_PREFIX + bone_name_list[15]]
    shin_R = bpy.data.objects[RD_PREFIX + bone_name_list[16]]
    foot_L = bpy.data.objects[RD_PREFIX + bone_name_list[17]]
    foot_R = bpy.data.objects[RD_PREFIX + bone_name_list[18]]
    toe_L = bpy.data.objects[RD_PREFIX + bone_name_list[19]]
    toe_R = bpy.data.objects[RD_PREFIX + bone_name_list[20]]   
    
    ##############
    # ADD JOINTS #  
    ##############    
    
    def create_rigid_body_constraint(obj, target, min_x, max_x, min_y, max_y, min_z, max_z):
        constraint = obj.constraints.new(type = 'RIGID_BODY_JOINT')
        constraint.name = 'RAGDOLL_PIVOT'
        constraint.target = target
        constraint.pivot_type = 'GENERIC_6_DOF'
        constraint.use_limit_x = True
        constraint.use_limit_y = True
        constraint.use_limit_z = True
        constraint.use_angular_limit_x = True
        constraint.use_angular_limit_y = True
        constraint.use_angular_limit_z = True
        constraint.limit_angle_min_x = radians(min_x)
        constraint.limit_angle_max_x = radians(max_x)
        constraint.limit_angle_min_y = radians(min_y)
        constraint.limit_angle_max_y = radians(max_y)
        constraint.limit_angle_min_z = radians(min_z)
        constraint.limit_angle_max_z = radians(max_z)
    
    #Create Rigid Body Constraints for Ragdoll Pieces    
    create_rigid_body_constraint(head, neck, -25, 20, -35, 35, -20, 20)
    create_rigid_body_constraint(neck, chest,  -20, 35, -15, 15, -20, 20)
    create_rigid_body_constraint(chest, spine, -35, 20, -20, 20, -15, 15)
    create_rigid_body_constraint(spine, hips, -20, 35, -15, 15, -15, 15)
    create_rigid_body_constraint(shoulder_L, chest, -5, 5, 0, 0, -10, 5)
    create_rigid_body_constraint(shoulder_R, chest, -5, 5, 0, 0, -5, 10)
    create_rigid_body_constraint(upper_arm_L, shoulder_L, -30, 60, -45, 35, -50, 60)
    create_rigid_body_constraint(upper_arm_R, shoulder_R, -30, 60, -45, 35, -60, 50)
    create_rigid_body_constraint(forearm_L, upper_arm_L, -10, 65, 0, 0, 0, 0)
    create_rigid_body_constraint(forearm_R, upper_arm_R, -10, 65, 0, 0, 0, 0)
    create_rigid_body_constraint(hand_L, forearm_L, -45, 30, 0, 0, -15, 15)
    create_rigid_body_constraint(hand_R, forearm_R, -45, 30, 0, 0, -15, 15)
    create_rigid_body_constraint(thigh_L, hips, -70, 45, -45, 15, -45, 30)
    create_rigid_body_constraint(thigh_R, hips, -70, 45, -15, 45, -30, 45)
    create_rigid_body_constraint(shin_L, thigh_L, -5, 70, 0, 0, 0, 0)
    create_rigid_body_constraint(shin_R, thigh_R, -5, 70, 0, 0, 0, 0)
    create_rigid_body_constraint(foot_L, shin_L, -20, 15, -5, 5, -5, 5)
    create_rigid_body_constraint(foot_R, shin_R, -20, 15, -5, 5, -5, 5)
    create_rigid_body_constraint(toe_L, foot_L, -15, 15, 0, 0, 0, 0)
    create_rigid_body_constraint(toe_R, foot_R, -15, 15, 0, 0, 0, 0)
    
    #Closing Statement:
    print("Ragdoll Generated")

#################################################################################################################################

def create_widget(rig, bone_name, PREFIX, where, make_parent):
    """ Creates an empty widget object for a bone and places it at the bone location (where)
    """
    obj_name = PREFIX + bone_name 

    # Check if it already exists in the scene
    if obj_name in bpy.context.scene.objects:
        return None
    else:
        # Delete object if it exists in blend data but not scene data.
        # This is necessary so we can then create the object without
        # name conflicts.
        if obj_name in bpy.data.objects:
            bpy.data.objects[obj_name].user_clear()
            bpy.data.objects.remove(bpy.data.objects[obj_name])
        # Create mesh object
        mesh = bpy.data.meshes.new(obj_name)
        obj = bpy.data.objects.new(obj_name, mesh)
        bpy.context.scene.objects.link(obj)
        # Move object to bone position
        obj_to_bone(obj, rig, bone_name, where, make_parent)
        obj.layers = [x == 18 for x in range(0, 20)]    
    return obj

def create_follow_widget(rig, bone_name, FOLLOW_PREFIX):
    """ Creates a cross widget
    """
    obj = create_widget(rig, bone_name, FOLLOW_PREFIX, 'head', True)
    
    if obj != None:
        verts =[(0.0, -0.1, 0.0), (0.0, 0.1, 0.0), (-0.0, -0.0, 0.1), (0.0, 0.0, -0.1), (0.1, 0.0, 0.0), (-0.1, -0.0, -0.0)]
        edges = [(0, 1), (2, 3), (4, 5)]
        mesh = obj.data
        mesh.from_pydata(verts, edges, [])
        mesh.update()

def create_ragdoll_widgets(rig, bone_name_list, FOLLOW_PREFIX, RD_PREFIX):
            
    #Create the Follow Widgets:
    for i in range(len(bone_name_list)):
        create_follow_widget(rig, bone_name_list[i], FOLLOW_PREFIX)
    print("Follow widgets created.")

def add_bone_constraints(rig, RD_PREFIX, bone_name_list):
    #Add IK constraints to each bone
    bpy.context.scene.objects.active = rig
    bpy.ops.object.mode_set(mode='POSE')
    for i in range(len(bone_name_list)):
        #Add limit rotation constraints
        pose_bone = bpy.data.objects[rig.name].pose.bones[bone_name_list[i]]
        pose_bone.constraints.new('COPY_ROTATION')
        pose_bone.constraints[len(pose_bone.constraints) - 1].name = 'RAGDOLL_ROTATION'
        pose_bone.constraints[len(pose_bone.constraints) - 1].target = bpy.data.objects[(RD_PREFIX + bone_name_list[i])]
        pose_bone.constraints[len(pose_bone.constraints) - 1].influence = 0
        #Add copy location constraints
        if bone_name_list[i] == 'hips':
            pose_bone.constraints.new('COPY_LOCATION')
            pose_bone.constraints[len(pose_bone.constraints) -1].name = 'RAGDOLL_LOCATION'
            pose_bone.constraints[len(pose_bone.constraints) -1].target = bpy.data.objects[RD_PREFIX + bone_name_list[i]]
            pose_bone.constraints[len(pose_bone.constraints) -1].influence = 0
        
            
    bpy.ops.object.mode_set(mode='OBJECT')
    print("COPY ROTATION, COPY LOCATION, AND LIMIT ROTATION CONSTRAINTS ADDED")
    
  

main()
