import bpy
from math import radians

def main():
    """
    This script is still experimental...
    Adds cube pieces to target bone locations, rotations, and scales.
    Adds rigid body joint constraints to the pieces.
    Adds Follow objects to the head of each of the target bones, parenting them to their corresponding bones
    Adds copy rotation constraints to the bones called 'RAGDOLL_ROTATION'
    Adds copy location constraint to the bone parented to a bone called 'root' or if root is not in the armature, adds the copy location constraint to the bone with no parent.
    Instructions: Select metarig, enter edit mode, select all the bones that will be part of the ragdoll, run script.

    You will need to model the cubes into a ragdoll and adjust the rigid body joint constraints yourself.
    """
    rig = bpy.context.scene.objects.active
    bone_seq = bpy.context.selected_editable_bones
    name_list = get_edit_info(bone_seq, "name")
    RD_PREFIX = 'RD-'
    FOLLOW_PREFIX = 'FOLLOW-'
    BODY_MASS = 100
    bpy.ops.object.mode_set(mode='OBJECT')
    
    create_ragdoll(rig, RD_PREFIX, BODY_MASS, name_list)
    create_ragdoll_widgets(rig, name_list, FOLLOW_PREFIX, RD_PREFIX)
    add_bone_constraints(rig, RD_PREFIX, name_list)
    #Make Relevant Layers visible 
    bpy.context.scene.layers = create_layers_boolean([(bpy.context.scene.active_layer + 1), 19], 20)
    
def get_edit_info(seq, target):
    """ While in edit mode, scans all bones in an editbone sequence 
        and returns a list of the target attribute e.g. head, bone length, etc.
    """
    attribute_list = []
    for i in range(len(seq)):
        attribute_list.append(getattr(seq[i], target))
    return attribute_list

def create_layers_boolean(list_of_layers, total_layers):
    """
    Given a list of layers and the total number of layers
    returns a boolean list of Trues and False
    """
    boolean_list = []
    for i in range(1, (total_layers + 1)):
        if i in list_of_layers:
            boolean_list.append(True)
        else:
            boolean_list.append(False)
    return boolean_list

def create_body_part(rig, bone_name, RD_PREFIX):
        """ Creates an empty object for a ragdoll body part, and returns the object.
        """
        obj_name = RD_PREFIX + bone_name
        scene = bpy.context.scene
        # Check if it already exists
        if obj_name in scene.objects:
            return None
        else:
            mesh = bpy.data.meshes.new(obj_name)
            obj = bpy.data.objects.new(obj_name, mesh)
            scene.objects.link(obj)
    
            obj_to_bone(obj, rig, bone_name, 'head', False)
    
            return obj
        
def obj_to_bone(obj, rig, bone_name, where, make_bone_parent):
        """ 
        Places an object at the location/rotation/scale of the given bone Warning! This script will clear
        whatever object was the parent of the obj before running the script.
        
        Parameters:     obj (bpy.types.Objects) the object that will be moved to the bone's position
                    
                        rig (bpy.types.Objects) the rig that has the bone to which the object will be moved
                        
                        bone_name (string) the name of the bone to which the object will be moved
                        
                        where (string) the location along the bone where the object should be placed 'head' 'tail' or 'middle' 
                        
                        make_bone_parent (boolean) Whether or not the object should be parented to the bone after being placed at
                        the bone's location.
        """       
        obj.select = True
        bpy.ops.object.parent_clear(type='CLEAR')
        bpy.ops.object.location_clear()
        bpy.ops.object.rotation_clear()
        bpy.ops.object.scale_clear()        
        obj.parent = rig
        obj.parent_type = 'BONE'
        obj.parent_bone = bone_name
        bone = rig.data.bones[bone_name]        
        if where == 'head':
            obj.location[1] -= bone.length
        elif where == 'middle':
            obj.location[1] -= bone.length/2        
        obj.rotation_euler = obj.parent.rotation_euler
        if make_bone_parent == True:
            obj.rotation_euler = obj.parent.rotation_euler
        else:
            bpy.ops.object.parent_clear(type='CLEAR_KEEP_TRANSFORM')
            mat = rig.matrix_world * bone.matrix_local
            obj.rotation_mode = ('XYZ')
            obj.rotation_euler = mat.to_euler()    
            scl = mat.to_scale()
            scl_avg = (scl[0] + scl[1] + scl[2]) / 3
            obj.scale = (bone.length * scl_avg), (bone.length * scl_avg), (bone.length * scl_avg)
        obj.select = False
        
#sets some standard things like physics type, mass etc. "ratio" is the proportion of mass given to each body part.
def standard_doll_settings(obj, ratio, name, BODY_MASS, RD_PREFIX):
    if obj != None:
        obj.game.physics_type = 'RIGID_BODY'
        obj.hide_render = False
        obj.game.mass = (ratio * BODY_MASS)
        obj.game.use_collision_bounds = True
        obj.game.collision_bounds_type = 'CONVEX_HULL'
        obj.select = True
        bpy.ops.object.transform_apply(location=False, rotation=False, scale=True)
        obj.select = False

    else:
        obj = bpy.data.objects[RD_PREFIX + name]
        
def create_ragdoll(rig, RD_PREFIX, BODY_MASS, name_list): 
    """ 
    Generates a ragdoll base using some of the bones of a metarig armature
        
    Parameters:     rig (bpy.data.objects) The armature object that will be recieving the ragdoll pieces
                    
                    RD_PREFIX (string) the prefix that will be added to each of the generated ragdoll
                    pieces
                    
                    layer (int between 1 and 20) the layer to which all of the ragdoll pieces will 
                    be added
                    
                    BODY_MASS (float) the total body mass of the ragdoll. Each piece is given a portion
    """
    
    def create_general_body_part(rig, bone_name, RD_PREFIX):
        obj = create_body_part(rig, bone_name, RD_PREFIX)
        if obj != None:
            verts = [(0.1, 0.1, -0.1), (0.1, -0.1, -0.1), (-0.1, -0.1, -0.1), (-0.1, 0.1, -0.1), (0.1, 0.1, 0.1), (0.1, -0.1, 0.1), (-0.1, -0.1, 0.1), (-0.1, 0.1, 0.1)]
            faces = [(0, 1, 2, 3), (4, 7, 6, 5), (0, 4, 5, 1), (1, 5, 6, 2), (2, 6, 7, 3), (4, 0, 3, 7)]
            mesh = obj.data
            mesh.from_pydata(verts, [], faces)
            mesh.update()
            obj.layers = [x == 17 for x in range(0, 20)] 
            return obj
        else:
            return None
        
    for i in range(len(name_list)):
        standard_doll_settings(create_general_body_part(rig, name_list[i], RD_PREFIX), .1, name_list[i], BODY_MASS, RD_PREFIX)
    ##############
    # ADD JOINTS #  
    ############## 
    
    def create_rigid_body_constraint(obj, target, min_x, max_x, min_y, max_y, min_z, max_z):
        constraint = obj.constraints.new(type = 'RIGID_BODY_JOINT')
        constraint.name = 'RAGDOLL_PIVOT'
        constraint.target = target
        constraint.pivot_type = 'GENERIC_6_DOF'
        constraint.use_limit_x = True
        constraint.use_limit_y = True
        constraint.use_limit_z = True
        constraint.use_angular_limit_x = True
        constraint.use_angular_limit_y = True
        constraint.use_angular_limit_z = True
        constraint.limit_angle_min_x = radians(min_x)
        constraint.limit_angle_max_x = radians(max_x)
        constraint.limit_angle_min_y = radians(min_y)
        constraint.limit_angle_max_y = radians(max_y)
        constraint.limit_angle_min_z = radians(min_z)
        constraint.limit_angle_max_z = radians(max_z)
        
    for i in range(len(name_list)):
        obj = bpy.data.objects[RD_PREFIX + name_list[i]]
        if rig.data.bones[name_list[i]].parent != None and rig.data.bones[name_list[i]].parent.name != 'root':
            create_rigid_body_constraint(obj, bpy.data.objects[RD_PREFIX + rig.data.bones[name_list[i]].parent.name], 0, 0, 0, 0, 0, 0)
            
def create_widget(rig, bone_name, PREFIX, where, make_parent):
    """ Creates an empty widget object for a bone and places it at the bone location (where)
    """
    obj_name = PREFIX + bone_name 

    # Check if it already exists in the scene
    if obj_name in bpy.context.scene.objects:
        return None
    else:
        # Delete object if it exists in blend data but not scene data.
        # This is necessary so we can then create the object without
        # name conflicts.
        if obj_name in bpy.data.objects:
            bpy.data.objects[obj_name].user_clear()
            bpy.data.objects.remove(bpy.data.objects[obj_name])
        # Create mesh object
        mesh = bpy.data.meshes.new(obj_name)
        obj = bpy.data.objects.new(obj_name, mesh)
        bpy.context.scene.objects.link(obj)
        # Move object to bone position
        obj_to_bone(obj, rig, bone_name, where, make_parent)
        obj.layers = [x == 18 for x in range(0, 20)]    
    return obj

def create_follow_widget(rig, bone_name, FOLLOW_PREFIX):
    """ Creates a cross widget
    """
    obj = create_widget(rig, bone_name, FOLLOW_PREFIX, 'head', True)
    
    if obj != None:
        verts =[(0.0, -0.1, 0.0), (0.0, 0.1, 0.0), (-0.0, -0.0, 0.1), (0.0, 0.0, -0.1), (0.1, 0.0, 0.0), (-0.1, -0.0, -0.0)]
        edges = [(0, 1), (2, 3), (4, 5)]
        mesh = obj.data
        mesh.from_pydata(verts, edges, [])
        mesh.update()

def create_ragdoll_widgets(rig, name_list, FOLLOW_PREFIX, RD_PREFIX):
            
    #Create the Follow Widgets:
    for i in range(len(name_list)):
        create_follow_widget(rig, name_list[i], FOLLOW_PREFIX)
    print("Follow widgets created.") 
    
def add_bone_constraints(rig, RD_PREFIX, name_list):
    #Add IK constraints to each bone
    bpy.context.scene.objects.active = rig
    bpy.ops.object.mode_set(mode='POSE')
    for i in range(len(name_list)):
        #Add limit rotation constraints
        pose_bone = bpy.data.objects[rig.name].pose.bones[name_list[i]]
        pose_bone.constraints.new('COPY_ROTATION')
        pose_bone.constraints[len(pose_bone.constraints) - 1].name = 'RAGDOLL_ROTATION'
        pose_bone.constraints[len(pose_bone.constraints) - 1].target = bpy.data.objects[(RD_PREFIX + name_list[i])]
        pose_bone.constraints[len(pose_bone.constraints) - 1].influence = 0
        if rig.data.bones[name_list[i]].parent == None or rig.data.bones[name_list[i]].parent.name == 'root':
            #Add copy location constraints
            pose_bone.constraints.new('COPY_LOCATION')
            pose_bone.constraints[len(pose_bone.constraints) -1].name = 'RAGDOLL_LOCATION'
            pose_bone.constraints[len(pose_bone.constraints) -1].target = bpy.data.objects[RD_PREFIX + name_list[i]]
            pose_bone.constraints[len(pose_bone.constraints) -1].influence = 0
        
    bpy.ops.object.mode_set(mode='OBJECT')
    print("COPY ROTATION, COPY LOCATION, AND LIMIT ROTATION CONSTRAINTS ADDED") 
    
main()
